#!/bin/sh

echo -e "\e[33m=== tools-umlsetup.sh" &&

# This script is run by the UML instance to set up the UML environment and
# pass control on to the mktools-build script.  If you don't need
# User Mode Linux (I.E. you're running as root on a new enough kerenel), you
# can theoretically skip this script and run the mktools-build script directly.

# So what does this setup entail?  Well, we start with the parent filesystem's
# "/" hostfs mounted on our "/".  This is mounted read-write, but not quite
# read-write enough.  All sorts of permissions are wrong.

# Although we're running as root inside UML, the user running UML on the host
# system probably isn't (shouldn't be) root.  As a result, large chunks of our
# hostfs are effectively read only.  Specifically, we can't write to the root
# directory to create /tools for mktools-build.sh, we can't mknod or chown
# anywhere, and on top of that the mounted synthetic filesystems (/proc
# and /dev/pts) refer to the parent kernel rather than the UML instance
# running our processes.

# To get around all this, we create a big ext2 image (512 megs) and loopback
# mount it.  This gives us writeable space with full root access in which to
# build.  We still need the parent system's build tools, so we bind mount all
# the top level directories (except proc) from the host's "/" into our ext2
# mount (and duplicate all the symlinks, so we can rebuild under ourselves),
# mount the UML instance's proc and devpts into it, and chroot.

# This assumes we have a sane host system, where /home, /proc, and /dev/pts
# exist, and where /mktools.sh and /tools do not already exist.  Then again,
# we're assuming you've got a sane enough host system to compile the rest of
# the tools directory anyway.  (If you haven't, just grab a prepackaged tools
# and use that.)

# Note that "sane" is not the same thing as "well-constructed".  For example,
# we use mount -n in case somebody is crazy enough to run this as root on an
# old system where /etc/mtab is not a symlink to /proc/mounts.  (Otherwise
# it'll corrupt /etc/mtab with mount points the host system doesn't have.
# A reboot should fix that if it happens, but still.)

mount -n -t proc /proc /proc &&
export PATH=/sbin:/usr/sbin:/bin:/usr/bin &&
cd "$DIR/tmpdir" &&
echo "=== Loopback mount workspace.img." &&
losetup /dev/loop0 workspace.img &&
mount -n -t ext2 /dev/loop0 sub &&
cd sub

if [ $? -ne 0 ]; then exit 1; fi

echo "=== Bind parent system's top level directories into loopback mount." &&
for i in /*
do
  i="${i:1}"
  if [[ ( "$i" != "lost+found" ) && ( "$i" != tmp ) && ( "$i" != sys ) ]]
  then
    echo -n "$i",
    if [ -h "$i" ]
    then
      # Copy symlinks
      ln -s `ls -l "$i" | sed 's/^.*-> //'` "$i"
    elif [ -d "/$i" ]
    then
      # Bind mount directories
      mkdir "$i" &&
      mount -n -o bind "/$i" "$i"
    fi
  fi
  if [ $? -ne 0 ]; then exit 1; fi
done
echo

# Don't use system /tmp, use a tmp in workspace.img.
mkdir tmp &&
# Mount sys.  Not guaranteed to be there on host system, and should use UML
# version anyway.
mkdir sys &&
mount -t sysfs sys sys &&
mknod -m 700 tmp/ubda b `sed 's/:.*//' sys/block/ubda/dev` \
	`sed 's/.*://' sys/block/ubda/dev` &&
swapon tmp/ubda &&

mount -n -t devpts /dev/pts dev/pts &&

# Setup swap

echo "=== chroot and hand control off to build script." &&
chroot . /bin/sh -c 'cd "$DIR" ; exec sources/scripts/1.1-tools-build.sh'

# Make a squashfs image of tools if the build script finished.

rm tools/finished 2> /dev/null &&
../mksquashfs tools ../tools.img -info

halt -f
