#!/tools/bin/sh

# This script is run in the chroot environment, with /tools and /sources in
# an otherwise empty (writeable) directory.

set +h
umask 022
unset CC CXX CPP LD_LIBRARY_PATH LD_PRELOAD
export LC_ALL=POSIX
export PATH=/sbin:/bin:/tools/bin
export HOME=/root
export TERM=linux

# Read config

if [ -f /sources/config ]
then
  source /sources/config
fi

# Autodetect architecture.
#export ARCH=`gcc -v 2>&1 | sed -n "s/.*--host=//p" | sed "s/-.*//"`
export ARCH=i386

echo === Make directories

# I haven't done this yet, but this is the plan:
# This is a bit idiosyncratic: some redundant directories have been merged,
# with symlinks from the standard positions pointing to their new
# positions.

# The set "bin->usr/bin, sbin->usr/sbin, lib->usr/lib" all serve to consolidate
# all the executables under /usr.  This has a bunch of nice effects: making a
# a read-only run from CD filesystem easier to do, allowing du /usr to show
# the whole system size, allowing everything outside of there to be mounted
# noexec, and of course having just one place to look for everything.  (Normal
# executables are in /usr/bin.  Root only executables are in /usr/sbin.
# Libraries are in /usr/lib.)

# For those of you wondering why /bin and /usr/sbin were split in the first
# place,  the answer is it's because Ken Thompson and Dennis Ritchie ran out
# of space on the original 2.5 megabyte RK-05 disk pack their root partition
# lived on in 1971, and leaked the OS into their second RK-05 disk pack where
# the user home directories lived.  (/usr was what /home is today.)

# The real reason we kept it is tradition.  The execuse is that the root
# partition contains early boot stuff and /usr may get mounted later, but these
# days we use "initrd" and "initramfs" to handle that sort of thing.  The
# version skew issues of actually trying to mix and match different
# versions of /lib/libc.so.* living on a local hard drive with a /usr/bin/*
# from the network mount are not pretty.

# I.E. The seperation is just a historical relic, and I've consolidated it in
# the name of simplicity.

# The one bit where this can cause a problem is merging /lib with /usr/lib,
# means that the same library can show up in the search path twice, and when
# that happens binutils gets confused and bloats the resulting executables.
# (They become as big as statically linked, but still refuse to run without
# opening the shared libraries.) The fix is to take /lib out of the binutils
# search path, which we do.  The symlink is left there in case somebody's using
# dlopen, and for standards compliance.

# Similarly, all the editable stuff has been moved under "var", including
# tmp->var/tmp, and etc->var/etc.  (Whether /etc really needs to be editable is
# an issue to be revisited later...)  Remember to move root's home directory
# somewhere writeable (I.E. /root should move to either /var/root or
# /home/root).

# The result is that a running system can have / be mounted read only (with
# /usr living under that), /var can be ramfs/tmpfs with a tarball extracted
# into it, /dev can be ramfs/tmpfs managed by udev (with /dev/pts being
# devpts), /proc can be procfs, /sys can bs sysfs.  Optionally, /home can be
# be an actual writeable filesystem on a hard drive or the network, and
# /dev/shm should be tmpfs if you're going to use shared memory.

# Make directory hierarchy with /{bin,sbin,lib} symlinked to /usr/{bin,sbin,lib}

mkdir -p {boot,dev/{pts,shm},home,mnt,proc,sys} &&
mkdir -p usr/{bin,etc,include,lib,sbin,src,share/{dict,doc,info,locale,man/man{1,2,3,4,5,6,7,8},nls,misc,terminfo,zoneinfo}} &&
ln -s share/{man,doc,info} usr &&
mkdir -p var/{lock,log,mail,run,spool,tmp,opt,cache,lib/misc,local} &&
mkdir -p opt/{bin,doc,include,info,lib,man/man{1,2,3,4,5,6,7,8}} &&
chmod 1777 var/tmp &&
# Collapse together some redundant directories (nonstandard, but compatable.)
ln -s .. usr/local &&
mkdir -p var/{boot,etc/opt} &&
ln -s var/tmp tmp &&
ln -s var/boot boot &&
ln -s var/etc etc &&
ln -s usr/bin bin &&
ln -s usr/sbin sbin &&
#ln -s ../var/tmp usr/tmp &&
# Temporary symlinks (should all be removed/overwritten during build)
ln -s /tools/lib lib &&
ln -s /tools/bin/{bash,cat,pwd,stty} /bin &&
ln -s /tools/bin/perl /usr/bin &&
ln -s bash /bin/sh &&
mount proc /proc -t proc &&
mount sysfs /sys -t sysfs &&
mount devpts /dev/pts -t devpts &&

if [ $? -ne 0 ]; then exit 1; fi

echo === Create passwd and group files.
cat>/etc/passwd <<EOF &&
root:x:0:0:root:/root:/bin/bash
EOF
cat>/etc/group<<EOF &&
root:x:0:
bin:x:1:
sys:x:2:
kmem:x:3:
tty:x:4:
tape:x:5:
daemon:x:6:
floppy:x:7:
disk:x:8:
lp:x:9:
dialout:x:10:
audio:x:11:
EOF
ln -s /proc/mounts /etc/mtab &&
echo CST6EDT > /etc/TZ

if [ $? -ne 0 ]; then exit 1; fi

echo "=== Makedev (replace with udev in 2.6)."

cd /dev &&
bzcat /sources/system/MAKEDEV-1.7.bz2 > MAKEDEV-1.7 &&
chmod +x MAKEDEV-1.7 &&
./MAKEDEV-1.7 -v generic-nopty &&
rm MAKEDEV-1.7

if [ $? -ne 0 ]; then exit 1; fi

echo === Unpack linux kernel headers somewhere uclibc can find them.

cd /tmp &&
tar xvjf /sources/buildtools/linux-libc-headers-2.6.12.0.tar.bz2 &&
cd linux-libc-headers-2.6.12.0/include &&
mv asm-$ARCH ../../asm &&
mv linux sound ../.. &&
cd ../.. &&
rm -rf linux-libc-headers-2.6.12.0

if [ $? -ne 0 ]; then exit 1; fi

echo === Install uclibc

cd /tmp &&
tar xvjf /sources/system/uClibc-0.9.28.tar.bz2 &&
cd uClibc-0.9.28 &&
bzcat /sources/system/uClibc-config.bz2 > .config &&
# fix_includes gets deeply confused by pre-prepared kernel headers, so
# do its job manually.
echo > extra/scripts/fix_includes.sh &&
mv ../asm ../linux ../sound include &&
make &&
rm /lib &&
ln -s /usr/lib /lib &&
ln -s /tools/lib/libgcc_s.so.1 /usr/lib &&
make PREFIX=/usr/ install_runtime &&
make install_dev &&
# Also install ldd and friends
make CFLAGS=-static install_utils &&
cd .. &&
rm -rf uClibc-0.9.28

if [ $? -ne 0 ]; then exit 1; fi

if [ ! -z $DO_BUILDTOOLS ]
then

echo === Binutils

cd /tmp &&
tar xvjf /sources/buildtools/binutils-2.16.1.tar.bz2 &&
sed -i -e 's@^NATIVE_LIB_DIRS.*@NATIVE_LIB_DIRS=/usr/lib@' binutils-2.16.1/ld/configure.tgt &&
mkdir binutils-build &&
cd binutils-build &&
../binutils-2.16.1/configure --prefix=/usr --enable-shared --disable-nls &&
make tooldir=/usr &&
make tooldir=/usr install &&
cp ../binutils-2.16.1/include/libiberty.h /usr/include &&
cd .. &&
rm -rf binutils-build binutils-2.16.1

if [ $? -ne 0 ]; then exit 1; fi

echo === gcc

cd /tmp &&
tar xvjf /sources/buildtools/gcc-core-4.0.2.tar.bz2 &&
#tar xvjf buildtools/gcc-g++-4.0.2.tar.bz2 &&
cd gcc-4.0.2 &&
# Do NOT run fixincludes (Replaces no_fixincludes patch from LFS)
sed -i -e 's@\./fixinc\.sh@-c true@' -e 's@.*include/README@	echo@' \
        gcc/Makefile.in &&
# Don't install libiberty, binutils already did.
sed -i "s/install:.*install-subdir/install: install-subdir/" libiberty/Makefile.in &&
# Adjust config entries from which spec file will be generated
for i in `find gcc/config -name "*.h"`
do
  sed -i -r -e 's /lib(64)?/ld(64)?(-linux)?(-x86-)?(-ia)?(64)?.so.[12] /lib/ld-uClibc.so.0 ' "$i"
  if [ $? -ne 0 ]; then exit 1; fi
done &&
mkdir ../gcc-build &&
cd ../gcc-build &&
../gcc-4.0.2/configure --prefix=/usr --enable-shared --enable-threads=posix --disable-nls --enable-languages=c && # --enable-__cxa_atexit --enable-clocale=gnu
make &&
make install &&
ln -s gcc /usr/bin/cc &&
cd .. &&
rm -rf gcc-4.0.2 gcc-build

if [ $? -ne 0 ]; then exit 1; fi

echo === make

cd /tmp &&
tar xvjf /sources/buildtools/make-3.80.tar.bz2 &&
cd make-3.80 &&
./configure --prefix=/usr &&
make &&
make install &&
cd .. &&
rm -rf make-3.80

if [ $? -ne 0 ]; then exit 1; fi

else
# If we have no buildtools, we don't need the include directory.
rm -rf /usr/include
fi # DO_BUILDTOOLS

# Base operating system

echo === busybox-1.1.0-pre1

cd /tmp &&
tar xvjf /sources/system/busybox-20051018.tar.bz2 &&
cd busybox &&
bzcat /sources/system/busybox-config.bz2 > .config &&
make &&
cp busybox /bin &&
make PREFIX=/ install &&
# Make stripped down busybox for initramfs
make allnoconfig &&
cat >> .config << EOF &&
CONFIG_STATIC=y
CONFIG_DD=y
CONFIG_HEAD=y
CONFIG_OD=y
CONFIG_TEST=y
CONFIG_READLINK=y
CONFIG_AWK=y
CONFIG_FEATURE_SH_IS_MSH=y
CONFIG_MSH=y
CONFIG_FEATURE_SH_STANDALONE_SHELL=y
CONFIG_LOSETUP=y
CONFIG_CHROOT=y
CONFIG_MOUNT=y
EOF
yes "" | make oldconfig &&
make clean &&
make &&
mv busybox ../initramfs-busybox &&
cd .. &&
rm -rf busybox

if [ $? -ne 0 ]; then exit 1; fi

# Stuff that busybox should be upgraded to replace...

echo === bash

cd /tmp &&
tar xvzf /sources/temp/bash-2.05b.tar.gz &&
cd bash-2.05b &&
patch -p1 -i /sources/temp/bash-2.05b-2.patch &&
# Yes, gcc 4.0.2 broke bash by gratuitously treating as an error something
# that should only be a warning.  Don't ask me why bash has its' own malloc
# implementation anyway.  Welcome to dealing with fsf code...
sed '/free_return:/a do {;} while(0);' -i lib/malloc/malloc.c &&
./configure --prefix=/usr --bindir=/bin &&
make &&
make install &&
cd .. &&
rm -rf bash-2.05b

if [ $? -ne 0 ]; then exit 1; fi

echo === diffutils

cd /tmp &&
tar xvzf /sources/temp/diffutils-2.8.1.tar.gz &&
cd diffutils-2.8.1 &&
./configure --prefix=/usr &&
make &&
make install &&
cd .. &&
rm -rf diffutils-2.8.1

if [ $? -ne 0 ]; then exit 1; fi

if [ ! -z $DO_MORE_BUILDTOOLS ]
then

echo === m4

cd /tmp &&
tar xvjf /sources/buildtools/m4-1.4.3.tar.bz2 &&
cd m4-1.4.3 &&
./configure --prefix=/usr &&
make &&
make install &&
cd .. &&
rm -rf m4-1.4.3

if [ $? -ne 0 ]; then exit 1; fi

echo === bison

cd /tmp &&
tar xvjf /sources/buildtools/bison-2.1.tar.bz2 &&
cd bison-2.1 &&
# As usual, the FSF managed to break stuff.  Bison compiles fine without c++
# but the examples don't, and there's no easy way to disable them...
mv examples examples.bak &&
mkdir -p examples/calc++ &&
echo -e "all:\ninstall:\nclean:" | tee examples/Makefile.in \
	> examples/calc++/Makefile.in &&
./configure --prefix=/usr &&
make &&
make install &&
cd .. &&
rm -rf bison-2.1

if [ $? -ne 0 ]; then exit 1; fi

echo === flex

cd /tmp &&
tar xvzf /sources/buildtools/flex-2.5.4a.tar.gz &&
cd flex-2.5.4 &&
./configure --prefix=/usr &&
make &&
make install &&
ln -s libfl.a /usr/lib/libl.a &&
cat>/usr/bin/lex<<EOF &&
#!/bin/sh
exec /usr/bin/flex -l "$@"
EOF
chmod 755 /usr/bin/lex &&
cd .. &&
rm -rf flex-2.5.4

if [ $? -ne 0 ]; then exit 1; fi

fi # DO_MORE_BUILDTOOLS

if [ ! -z $DO_BOOTLOADER ]
then

# Bootloader

echo "=== bin86 (needed by lilo)"

cd /tmp &&
tar xvzf /sources/buildtools/bin86-0.16.17.tar.gz &&
cd bin86-0.16.17 &&
sed -i "s@^PREFIX=.*@PREFIX=/usr@" Makefile &&
make &&
make install &&
cd .. &&
rm -rf bin86-0.16.17

if [ $? -ne 0 ]; then exit 1; fi

echo "=== as86 (also needed by lilo)"

cd /tmp &&
tar xvzf /sources/buildtools/as86-0.16.17.tar.gz &&
cd as86-0.16.17 &&
make &&
make install &&
cd .. &&
rm -rf as86-0.16.17

if [ $? -ne 0 ]; then exit 1; fi

echo === lilo-22.6.1

cd /tmp &&
tar xvzf /sources/system/lilo-22.7.1.src.tar.gz &&
cd lilo-22.7.1 &&
# My tweak for a "length" option
bzcat ../system/lilo-length.patch.bz2 | patch -p1
make &&
make install &&
# A basic lilo.conf
cat > /etc/lilo.conf << EOF &&
boot=/dev/hda
compact
delay=10
timeout=1

image=/boot/vmlinuz
	label=linux
	root=/dev/loop0
	initrd=/initrd.img.gz
	append="init=/sbin/init"
	read-write
EOF
cd .. &&
rm -rf lilo-22.7.1

if [ $? -ne 0 ]; then exit 1; fi

fi # DO_BOOTLOADER

#echo === zlib
#
#cd /tmp &&
#tar xvjf /sources/applibs/zlib-1.2.3.tar.bz2 &&
#cd zlib-1.2.3 &&
#./configure --prefix=/usr --shared &&
#make &&
#make install &&
#make clean &&
#./configure --prefix=/usr &&
#make &&
#make install &&
#cp zlib.3 /usr/share/man/man3 &&
#cd .. &&
#rm -rf zlib-1.2.3
#
#if [ $? -ne 0 ]; then exit 1; fi

# And now, some optional applications.

#echo "=== dropbear-0.46 (ssh server and client)"
#
#cd /sources &&
#tar xvjf apps/dropbear-0.46.tar.bz2 &&
#cd dropbear-0.46 &&
## Switch off some things #switch others on.
#sed -ir -e "s/#define (DROPBEAR_SMALL_CODE|DO_HOST_LOOKUP|DO_MOTD)//" options.h
#./configure --prefix=/usr &&
#make PROGRAMS="dropbear dbclient dropbearkey scp" MULTI=1 &&
#strip dropbearmulti &&
#mv dropbearmulti /bin &&
#ln -s /bin/dropbearmulti /bin/ssh /bin/scp &&
#ln -s /bin/dropbearmulti /sbin/dropbear /sbin/dropbearkey &&
## Create RSA host key.
#mkdir /etc/dropbear &&
#dropbearkey -t rsa -f /etc/dropbear/dropbear_rsa_host_key \
#	-y /etc/dropbear/dropbear_rsk_host_key.pub &&
#cd .. &&
#rm -rf dropbear-0.46
#
#if [ $? -ne 0 ]; then exit 1; fi

# At this point, we've build a system you can chroot into, but not
# actually boot into.  We need a kernel, lilo, and so on.

#echo === linux-2.6.13.2
#
#cd /sources &&
#tar xvjf system/linux-2.6.13.2.tar.bz2 &&
#cd linux-2.6.13.2 &&
#bzcat ../system/linux-config.bz2 > .config &&
#yes "" | make oldconfig &&
#make &&
#cp System.map /boot &&
#cp arch/i386/boot/bzImage /boot/vmlinuz-2.6.13.2 &&
#ln -s vmlinuz-2.6.13.2 /boot/bzImage &&
#cd .. &&
#rm -rf linux-2.6.13.2
#
#if [ $? -ne 0 ]; then exit 1; fi

# Now that that's in there, we need some tools to make a firmware image
# with.  Our firmware image is a loopback mounted zisofs image, so we
# need mkisofs and mkzftree.

#cd /sources &&
#tar xvjf buildtools/cdrtools-2.00.3.tar.bz2 &&
#cd cdrtools-2.00.3 &&
#make &&
# The install of this is somewhat broken, and most of the stuff it compiled
# is garbage anyway.  Just install what we want by hand.  (No need
# copy libraries, it's all statically linked.  It's evil, really.)
#cp mkisofs/OBJ/*/mkisofs cdrecord/OBJ/*/cdrecord cdda2wav/OBJ/*/cdda2wav /usr/bin &&
#cp mkisofs/mkisofs.8 /usr/man/man8 &&
#cp cdrecord/cdrecord.1 cdda2wav/cdda2wav.1 /usr/man/man1 &&
#cd .. &&
#rm -rf cdrtools-2.00.3
#
#if [ $? -ne 0 ]; then exit 1; fi

#install -m 644 /sources/protocols /etc/protocols &&
#install -m 644 /sources/services /etc/services
#
#if [ $? -ne 0 ]; then exit 1; fi


# Note: better /etc/inittab
# /etc/hosts, /etc/resolv.conf
# linux kernel
# grub

# /tools/bin/find /{,usr/,usr/local/}{bin,sbin,lib} -type f -exec /tools/bin/strip --strip-debug '{}' ';

# That doesn't work (find incomplete).  This does:
for i in `find /usr/{bin,sbin} -type f`; do strip --strip-unneeded $i; done
for i in `find /{,usr/}lib -type f`; do strip --strip-debug $i; done

# rm -rf tools/{doc,info,man}
umount /proc /sys /dev/pts &&
touch /build_complete
